# Certificates and Witnesses for Probabilistic Model Checking -- Supplementary material 
This repo includes all data and code that was used to run the experiments in the thesis "Certificates and Witnesses for Probabilistic Model Checking".

## Preconditions

- Make sure the python package `virtualenv` is installed.
- To fully reproduce experiments an installation of [`Gurobi`](https://www.gurobi.com/) is required. Gurobi will be installed automatically by the setup script, but an active licence needs to be available on the machine to use it. However, you can reproduce almost all results without using CBC. To this end, adapt the variables `lp_solvers` and `milp_solvers` in the file `benchmarks/config.py` by replacing "gurobi" by "cbc".
- To reproduce the experiments of Chapter 5 (Probabilistic systems with low tree width), an installation of [`graph-tool`](https://graph-tool.skewed.de/) in the local python is required. To enable it, set `install_with_treealgo` to true in `setup.sh`.
- To reproduce the experiments involving the tool comics, it has to be installed first. It is available [here](https://www-i2.informatik.rwth-aachen.de/i2/comics/). After installing the tool, set the variable `comics_path` to contain the path to the binary of comics.

## Reproducing the experimental results

1. Adapt the variable `PYTHON` in `setup.sh` to link to the python installation you want to use.
2. Run `setup.sh`. A local python environment (using `virtualenv`) will be created in `switss_local_env`.
3. Source the virtual python environment (`source switss_local_env/bin/activate`) and then run any of the benchmark scripts named `bench_*` in the folder `benchmarks` (e.g., `python benchmarks/bench_switss_main.py`). Alternatively, run `run.sh`.
4. Run `python eval/evaluate_data.py` to compile plots and tables. This has to be done after activating the virtual python environment in `switss_local_env`.

## Structure
- `benchmarks`: contains python files configuring the benchmarks and python scripts to run them.
  - `benchmarks/benchmarks.py` specifies the benchmark instances.
  - `benchmarks/config.py` specifies various global configurations such as the timeout, and which solvers are used.
  - `comics_helpers.py` includes functions for interfacing with comics.
  - the files `bench_*.py` include scripts to run the experiments.
- `setup.sh`: script that sets up the environment for the experiments. It will
  - download the version of SWITSS used to run the experiments in the paper.
  - download the benchmark from the prism-benchmark suite.
  - install a local python environment which can be used to run the experiments
- `run.sh`: runs the experiments under the defined resource constraints.
- `eval/evaluate_data.py`: produces the figures and tables from the paper, and puts them into `plots/` and `tables/`
- `data-final`: contains all output files used in the paper. When running the experiments, the corresponding files will be put into a new directory called `data`.
- `logs`: contains log and debugging outputs (everything written to stdout during the experiment runs). This folder will be populated anew when running the experiments.
- `switss`: contains the version of switss used to run the experiments (this will be downloaded anew when running `setup.sh`)
- `prism-benchmarks`: contains the benchmark suite of PRISM (this will be downloaded anew when running `setup.sh`)
