import os
import pathlib
import time as time
import json
import tempfile
import sys

from switss import benchmarks as bm
from switss.model import DTMC, MDP, ReachabilityForm
from switss.problem import QSHeur, MILPExact, AllOnesInitializer

from config import *
from benchmarks import *

for bench in only_brp:
    for model_c in [(8,8),(32,8),(128,8),(1024,8)]:

        N,K = model_c
        extra_labels = {("success","s=4 & i=" + str(N))}
        labels_for_minimization = []
        for i in range(K):
            extra_labels.add(("nrtr_" + str(i),"nrtr=" + str(i)))
            labels_for_minimization.append("nrtr_" + str(i))

        start_load = time.perf_counter()
        M = DTMC.from_prism_model(
            bench["prism_file"](model_c),
            prism_constants = bench["prism_constants"](model_c),
            extra_labels = extra_labels)
        time_load = time.perf_counter() - start_load

        init_label = bench["init_label"]
        goal_label = "success"

        start_rf_constr = time.perf_counter()
        rf,_,_ = ReachabilityForm.reduce(M,init_label,goal_label)
        time_rf_constr = time.perf_counter() - start_rf_constr

        # with open("test.dot","w+") as testfile:
        #     testfile.write(str(rf.system.digraph()))

        json_dir = Path ("./data/brp_labeled") / ("brp.%d.%d" % (N,K))
        json_dir.mkdir(parents=True,exist_ok=True)

        info_dict = {
            "rf" : str(rf),
            "nr_states" : rf.A.shape[1],
            "nr_sap" : rf.A.shape[0],
            "ecfree" : bool(rf.is_ec_free),
            "time_loadmodel" : time_load,
            "time_reduce" : time_rf_constr
        }

        info_json = json.dumps(info_dict)

        with open(json_dir / "info.json","w") as infofile:
            infofile.write(info_json)

        instances = [ QSHeur(iterations=5,initializertype=AllOnesInitializer,solver="gurobi"),
                      MILPExact("gurobi") ]

        bm.run(rf,
               instances,
               ["max","min"],
               from_thr=0.97,
               step=0.002,
               debug=True,
               json_dir=json_dir,
               timeout=timeout,
               labels=labels_for_minimization,
               stop_on_timeout=False)
