import os
import pathlib
import tempfile

from switss.model import DTMC, ReachabilityForm
from switss.certification import generate_farkas_certificate

from config import *
from benchmarks import *
from comics_helpers import run_comics, prism_to_comics, write_header


if comics_path is None:
    print("not running comics experiments as comics_path not set!")

for bench in only_dtmcs:
    for model_c in bench["model_configurations"]:

        # create switss-model, to check feasibility below
        M = DTMC.from_prism_model(
            bench["prism_file"](model_c),
            prism_constants = bench["prism_constants"](model_c),
            extra_labels = bench["extra_labels"](model_c))

        init_label = bench["init_label"]
        goal_label = bench["goal_label"]

        rf,_,_ = ReachabilityForm.reduce(M,init_label,goal_label)

        pr = rf.pr_min()

        with tempfile.NamedTemporaryFile(suffix=".dtmc") as temp_comics_model_file:

            #write comics description of DTMC to tempfile
            prism_to_comics(bench["prism_file"](model_c),
                            bench["goal_label"],
                            temp_comics_model_file.name,
                            bench["prism_constants"](model_c),
                            bench["extra_labels"](model_c))

            thr = bench["from_thr"]
            json_dir_main = bench["json_dir_main"]

            json_dir_inst = Path(bench["json_dir_inst"](model_c))
            json_dir_inst.mkdir(parents=True,exist_ok=True)

            comics_csv_file = bench["json_dir_inst"](model_c) / "comics-run.csv"
            write_header(comics_csv_file)

            while thr <= pr[rf.initial]:
                run_comics(temp_comics_model_file.name, comics_csv_file, thr)
                thr += bench["thr_step"]
