import os
import pathlib
import time as time
import json
import tempfile
import sys

from switss import benchmarks as bm
from switss.model import DTMC, MDP, ReachabilityForm

from config import *
from benchmarks import *
from comics_helpers import run_comics, prism_to_comics, write_header

timeout = 60 * 60

brp_cp = brp_configuration.copy()
brp_cp["model_configurations"] = [(32,2)]

crowds_cp = crowds_configuration.copy()
crowds_cp["model_configurations"] = [(2,8)]

consensus_cp = consensus_configuration.copy()
consensus_cp["model_configurations"] = [(2,4)]

firewire_cp = firewire_configuration.copy()
firewire_cp["model_configurations"] = [3]

benchmarks = [ brp_cp, crowds_cp, consensus_cp, firewire_cp]

for bench in benchmarks:
    for model_c in bench["model_configurations"]:

        start_load = time.perf_counter()
        model = {"MDP" : MDP, "DTMC" : DTMC}[bench["model_type"]]
        M = model.from_prism_model(
            bench["prism_file"](model_c),
            prism_constants = bench["prism_constants"](model_c),
            extra_labels = bench["extra_labels"](model_c))
        time_load = time.perf_counter() - start_load

        init_label = bench["init_label"]
        goal_label = bench["goal_label"]

        start_rf_constr = time.perf_counter()
        rf,_,_ = ReachabilityForm.reduce(M,init_label,goal_label)
        time_rf_constr = time.perf_counter() - start_rf_constr

        json_dir_main = bench["json_dir_main"]

        json_dir_inst = Path(bench["json_dir_inst"](model_c)) / "exact-longer-to"

        json_dir_inst.mkdir(parents=True,exist_ok=True)

        instances = exact_instances.copy()

        stop_on_timeout = False

        bm.run(rf,
               instances,
               ["max"],
               from_thr=bench["from_thr"],
               step=bench["thr_step"],
               json_dir=json_dir_inst,
               timeout=timeout,
               debug=debug,
               stop_on_timeout=stop_on_timeout)
