import os
import pathlib
import time as time
import json
import tempfile
import sys

from switss.solver import GurobiMILP
from switss import benchmarks as bm
from switss.model import DTMC, MDP, ReachabilityForm
from switss.test import *
from switss.certification import generate_farkas_certificate

from config import *
from benchmarks import *

def construct_ltlsubsys_milp(rf,thr):
    ltlsubsys_milp = GurobiMILP()

    C,N = rf.P.shape

    p_vars = ltlsubsys_milp.add_variables_w_bounds([("real",0,1)] * N)

    x_vars = ltlsubsys_milp.add_variables_w_bounds([("integer",0,1)] * N)

    sig_vars = ltlsubsys_milp.add_variables_w_bounds([("integer",0,1)] * C)

    ## p_s <= x_s (8d)
    for i in range(N):
        ltlsubsys_milp.add_constraint([(p_vars[i],1),(x_vars[i],-1)],"<=",0)

    state_to_acts_dict = {}
    for i in range(C):
        (s,a) = rf.index_by_state_action.inv[i]
        if s not in state_to_acts_dict:
            state_to_acts_dict[s] = [i]
        else:
            state_to_acts_dict[s].append(i)

    ## sum \sig_{s,a} = x_s (8e)
    for s in range(N):
        acts = state_to_acts_dict[s]
        x_sum = [(x_vars[s],1)]
        sig_sum = [(sig_vars[i],-1) for i in acts]
        ltlsubsys_milp.add_constraint(x_sum + sig_sum,"=",0)

    ## prob_condition (8f)
    for i in range(C):
        (s,a) = rf.index_by_state_action.inv[i]
        row = rf.P.getrow(i)
        lhs = [(p_vars[s],1),(sig_vars[i],1)]
        p_sum = [0]*len(row.indices)
        for i,j,d in zip(range(len(row.indices)), row.indices, row.data):
            p_sum[i] = (p_vars[j], -float(d))
        ltlsubsys_milp.add_constraint(lhs + p_sum,"<=",rf.to_target[i] + 1)

    ## instate >= lambda
    ltlsubsys_milp.add_constraint([(p_vars[rf.initial],1)],">=",thr)

    optim = [(x,1) for x in x_vars]
    ltlsubsys_milp.set_objective_function(optim)

    return ltlsubsys_milp


for bench in all_benchmarks:
    for model_c in bench["model_configurations"]:

        if bench["instances"](model_c) == heur_instances:
            continue

        model = {"MDP" : MDP, "DTMC" : DTMC}[bench["model_type"]]

        start_load = time.perf_counter()
        M = model.from_prism_model(
            bench["prism_file"](model_c),
            prism_constants = bench["prism_constants"](model_c),
            extra_labels = bench["extra_labels"](model_c))
        time_load = time.perf_counter() - start_load

        init_label = bench["init_label"]
        goal_label = bench["goal_label"]

        start_rf_constr = time.perf_counter()
        rf,_,_ = ReachabilityForm.reduce(M,init_label,goal_label)
        time_rf_constr = time.perf_counter() - start_rf_constr

        prmax = rf.pr_max()

        if not rf.is_ec_free:
            continue

        json_dir_main = bench["json_dir_main"]

        json_dir_inst = Path(bench["json_dir_inst"](model_c))
        json_dir_inst.mkdir(parents=True,exist_ok=True)

        json_file = bench["json_dir_inst"](model_c) / "ltlsubsys-run.json"
        data = { "method" : {"type" : "ltlsubsys"}, "mode" : "max", "run" : []}

        ## construct ltlsubsys MILP using gurobi
        thr = bench["from_thr"]
        while thr <= prmax[rf.initial]:
            ltlsubsys_milp = construct_ltlsubsys_milp(rf,thr)

            start_solve = time.perf_counter()
            solverres = ltlsubsys_milp.solve(solver="gurobi",timeout=timeout)
            time_solve = time.perf_counter() - start_solve

            if solverres.status == "optimal":
                els = { "threshold" : float(thr), "value" : solverres.value, "wall_times" : float(time_solve) }
                data["run"].append(els)

            thr += bench["thr_step"]

            data_json = json.dumps(data)
            with open(json_file,"w") as f:
                f.write(data_json)
