import os
import pathlib
import time as time
import json
import tempfile
import sys

from switss import benchmarks as bm
from switss.model import DTMC, MDP, ReachabilityForm

from config import *
from benchmarks import *
from comics_helpers import run_comics, prism_to_comics, write_header

if "--mdps" in sys.argv:
    benchmarks = only_mdps
elif "--dtmcs" in sys.argv:
    benchmarks = only_dtmcs
else:
    benchmarks = all_benchmarks

for bench in benchmarks:
    for model_c in bench["model_configurations"]:

        start_load = time.perf_counter()
        model = {"MDP" : MDP, "DTMC" : DTMC}[bench["model_type"]]
        M = model.from_prism_model(
            bench["prism_file"](model_c),
            prism_constants = bench["prism_constants"](model_c),
            extra_labels = bench["extra_labels"](model_c))
        time_load = time.perf_counter() - start_load

        init_label = bench["init_label"]
        goal_label = bench["goal_label"]

        start_rf_constr = time.perf_counter()
        rf,_,_ = ReachabilityForm.reduce(M,init_label,goal_label)
        time_rf_constr = time.perf_counter() - start_rf_constr

        json_dir_main = bench["json_dir_main"]

        json_dir_inst = Path(bench["json_dir_inst"](model_c))

        json_dir_inst.mkdir(parents=True,exist_ok=True)

        prmin = rf.pr_min()
        prmax = rf.pr_max()

        info_dict = {
            "rf" : str(rf),
            "prmin" : prmin[rf.initial],
            "prmax" : prmax[rf.initial],
            "nr_states" : rf.A.shape[1],
            "nr_sap" : rf.A.shape[0],
            "ecfree" : bool(rf.is_ec_free),
            "time_loadmodel" : time_load,
            "time_reduce" : time_rf_constr
        }

        info_json = json.dumps(info_dict)

        with open(json_dir_inst / "info.json","w") as infofile:
            infofile.write(info_json)

        instances = bench["instances"](model_c)

        stop_on_timeout = bench["stop_on_timeout"](model_c)

        bm.run(rf,
               instances,
               ["min","max"],
               from_thr=bench["from_thr"],
               step=bench["thr_step"],
               json_dir=json_dir_inst,
               timeout=timeout,
               debug=debug,
               stop_on_timeout=stop_on_timeout)
