import os
import pathlib
import time as time
import json
import tempfile
import sys

from switss import benchmarks as bm
from switss.model import DTMC, MDP, ReachabilityForm

from config import *
from benchmarks import *

try:
    from switss.utils.tree_decomp import min_witnesses_from_tree_decomp
    from switss.utils.graph_utils import underlying_graph, quotient
    from switss.problem import TreeAlgo
    run_treealgo = True
except:
    print("switss installed without --with-treealg. TreeAlgo benchmarks will be ommited")
    run_treealgo = False

# define tree partition used for brp benchmarks
def compute_brp_tree_partition(rf,labels,to_rf_state_map):

    first_partition = []
    for i in range(0,N+1):
        i_states = []
        for q in labels["i" + str(i)]:
            if q in to_rf_state_map.keys():
                i_states.append(to_rf_state_map[q])
        first_partition.append(i_states)

    G = underlying_graph(rf.P)
    decomp, labeling, interface = quotient(G,first_partition)
    new_partition = []
    tree_width = 0
    for p in decomp.get_vertices():
        p_sucs = decomp.get_out_neighbors(decomp.vertex(p))
        new_p_states = [int(q) for q in G.get_vertices()
                        if (labeling[q] == p) and (not interface[q]) or (labeling[q] in p_sucs and interface[p])]
        new_partition.append(new_p_states)
        tree_width = max(tree_width,len(new_p_states))
    return new_partition,tree_width

# collect instances to run from commandline 
N_vals_to_run = map(int, sys.argv[1].strip('[]').split(','))

for N in N_vals_to_run: #10,20,30,40,50,60,70,80,90]:
    for MAX in [1,2,4,8]:
        extr_labels = {("uncertain","s=5 & srep=2"), *((("i" + str(i), "i=" + str(i)) for i in range(0,N+1)))}
        M = DTMC.from_prism_model(
            "./prism-benchmarks/models/dtmcs/brp/brp.pm",
            prism_constants={("N",N),("MAX",MAX)},
            extra_labels=extr_labels)

        rf,to_rf_state_map,_ = ReachabilityForm.reduce(M,"init","uncertain")

        labels = M.states_by_label

        partition,width = compute_brp_tree_partition(rf,labels,to_rf_state_map)

        json_dir = Path("./data/tree-algo/") / ("brp.%d.%d" % (N,MAX))
        json_dir.mkdir(parents=True,exist_ok=True)

        prmin = rf.pr_min()
        prmax = rf.pr_max()

        info_dict = {
            "rf" : str(rf),
            "prmin" : prmin[rf.initial],
            "prmax" : prmax[rf.initial],
            "nr_states" : rf.A.shape[1],
            "partition" : partition,
            "partition_width" : width
        }

        info_json = json.dumps(info_dict)

        with open(json_dir / "info-treealgo.json","w") as infofile:
            infofile.write(info_json)

        # run the treealgo on the instance
        bm.run(rf,
               [TreeAlgo(partition)],
               ["max"],
               from_thr=10e-5,
               step=10e-5,
               json_dir=json_dir,
               timeout=timeout,
               debug=True)

        # run exact computations using MILPs for comparison
        bm.run(rf,
               [MILPExact("cbc"),MILPExact("gurobi")],
               ["max","min"],
               from_thr=10e-5,
               step=10e-5,
               json_dir=json_dir,
               timeout=timeout,
               debug=True)
