import itertools
from config import *

# DTMCs

leader_configuration = {
    "model_type" : "DTMC",
    "json_dir_main" : global_json_dir / "leader",
    "json_dir_inst" : (lambda t: global_json_dir / "leader" / ("leader_%d_%d" % t)),
    "model_configurations" :  [(3,2),(3,8),(5,8),(6,6)],
    "prism_file" : (lambda t : prism_dtmcs + "leader_sync/leader_sync" + str(t[0]) + "_" + str(t[1]) + ".pm"),
    "prism_constants" : (lambda _ : {}),
    "extra_labels" : (lambda _ : {}),
    "init_label" : "init",
    "goal_label" : "elected",
    "instances" : (lambda t: heur_instances.copy() + exact_instances.copy() if t[0] <= 5 or t[1] <= 2 else heur_instances.copy()),
    "from_thr" : 0.1,
    "thr_step" : 0.2,
    "stop_on_timeout" : (lambda _ : True)
}

brp_configuration = {
    "model_type" : "DTMC",
    "json_dir_main" : global_json_dir / "brp",
    "json_dir_inst" : (lambda t: global_json_dir / "brp" / ("brp_%d_%d" % t)),
    "model_configurations" : [(16,2),(32,2),(1024,2),(2048,2)],
    "prism_file" : (lambda _ : prism_dtmcs + "brp/brp.pm"),
    "prism_constants" : (lambda t : {("N",t[0]),("MAX",t[1])}),
    "extra_labels" : (lambda _ : {("uncertain","s=5 & srep=2")}),
    "init_label" : "init",
    "goal_label" : "uncertain",
    "instances" : (lambda t: heur_instances.copy() + exact_instances.copy() if t[0] <= 2000 else heur_instances.copy()),
    "from_thr" : 2e-6,
    "thr_step" : 2e-6,
    "stop_on_timeout" : (lambda _ : True)
}

crowds_configuration = {
    "model_type" : "DTMC",
    "json_dir_main" : global_json_dir / "crowds",
    "json_dir_inst" : (lambda t: global_json_dir / "crowds" / ("crowds_%d_%d" % t)),
    "model_configurations" : [(2,6),(2,8),(5,8)],
    "prism_file" : (lambda _ : prism_dtmcs + "crowds/crowds.pm"),
    "prism_constants" : (lambda t : {("CrowdSize",t[0]),("TotalRuns",t[1])}),
    "extra_labels" : (lambda _ : {("target","observe0>1")}),
    "init_label" : "init",
    "goal_label" : "target",
    "instances" : (lambda t : heur_instances.copy() + exact_instances.copy() if t[0] <= 2 else heur_instances.copy()),
    "from_thr" : 0.01,
    "thr_step" : 0.02,
    "stop_on_timeout" : (lambda _ : True)
}

# MDPs

consensus_configuration = {
    "model_type" : "MDP",
    "json_dir_main" : global_json_dir / "consensus",
    "json_dir_inst" : (lambda t: global_json_dir / "consensus" / ("consensus_%d_%d" % t)),
    "model_configurations" : [(2,4),(4,2),(4,4)],
    "prism_file" : (lambda t : prism_mdps + "consensus/coin" + str(t[0]) + ".nm"),
    "prism_constants" : (lambda t : {("K",t[1])}),
    "extra_labels" : (lambda _ : {}),
    "init_label" : "init",
    "goal_label" : "finished",
    "instances" : (lambda t : heur_instances.copy() + exact_instances.copy() if t[0] <= 2 else heur_instances.copy()),
    "from_thr" : 0.1,
    "thr_step" : 0.1,
    "stop_on_timeout" : (lambda t : False if t == (2,4) else True)
}

csma_configuration = {
    "model_type" : "MDP",
    "json_dir_main" : global_json_dir / "csma",
    "json_dir_inst" : (lambda t: global_json_dir / "csma" / ("csma_%d_%d" % t)),
    "model_configurations" : [(2,2),(2,6),(3,2)],
    "prism_file" : (lambda t : prism_mdps + "csma/csma" \
                               + str(t[0]) + "_" + str(t[1]) + ".nm"),
    "prism_constants" : (lambda _ : {}),
    "extra_labels" : (lambda _ : {}),
    "init_label" : "init",
    "goal_label" : "all_delivered",
    "instances" : (lambda t: heur_instances.copy()  + exact_instances.copy() if t[0] == 2 and t[1] <= 4 else heur_instances.copy()),
    "from_thr" : 0.1,
    "thr_step" : 0.1,
    "stop_on_timeout" : (lambda _ : True)
}

firewire_configuration = {
    "model_type" : "MDP",
    "json_dir_main" : global_json_dir / "firewire",
    "json_dir_inst" : (lambda d: global_json_dir / "firewire" / ("firewire_%d" % d)),
    "model_configurations" : [3,30,40],
    "prism_file" : (lambda _ : prism_mdps + "firewire/firewire.nm"),
    "prism_constants" : (lambda d : {("delay",d)}),
    "extra_labels" : (lambda _ : {}),
    "init_label" : "init",
    "goal_label" : "done",
    "instances" : (lambda d: heur_instances.copy() + exact_instances.copy() if d <= 10 else heur_instances.copy()),
    "from_thr" : 0.1,
    "thr_step" : 0.1,
    "stop_on_timeout" : (lambda d : False if d == 3 else True)
}

zeroconf_configuration = {
    "model_type" : "MDP",
    "json_dir_main" : global_json_dir / "zeroconf",
    "json_dir_inst" : (lambda t: global_json_dir / "zeroconf" / ("zeroconf_%s_%d_%d" % t)),
    "model_configurations" : [("false",20,2),("true",20,2),("true",1000,2),("false",1000,4)],
    "prism_file" : (lambda t : prism_mdps + "zeroconf/zeroconf.nm"),
    "prism_constants" : (lambda t : {("reset",t[0]),("N",t[1]),("K",t[2])}),
    "extra_labels" : (lambda _ : {("target","l=4 & ip=1")}),
    "init_label" : "init",
    "goal_label" : "target",
    "instances" : (lambda t: heur_instances.copy()  + exact_instances.copy() if t[0] == "true" else heur_instances.copy()),
    "from_thr" : 0.1,
    "thr_step" : 0.05,
    "stop_on_timeout" : (lambda _ : True)
}

only_brp = [brp_configuration]
only_crowds = [crowds_configuration]
only_leader = [leader_configuration]

only_csma = [csma_configuration]
only_consensus = [consensus_configuration]
only_firewire = [firewire_configuration]
only_zeroconf = [zeroconf_configuration]

only_dtmcs = only_leader + only_brp + only_crowds
only_mdps = only_csma + only_consensus + only_firewire

all_benchmarks = only_dtmcs + only_mdps
