import sys
#sys.path.insert(0, "../switss")
from pathlib import Path,PurePath

import json as json
import matplotlib.pyplot as plt
import itertools
import numpy as np
import csv
import math
import re

import switss.benchmarks as bm

data_folder = Path("./data-final")

def add_timeouts(thr_list,data):
    thr_in_data = [el["threshold"] for el in data["run"]]
    for thr in thr_list:
        if thr not in thr_in_data:
            data["run"].append({ "threshold" : thr, "value" : 5*[-1], "wall_times" : 5*[-1], "proc_times" : 5*[-1] })

    return data

def comics_to_proper_json(comics_file,max_thr):
    data_comics_global = { "method" : {"type" : "comics"}, "mode" : "global", "run" : []}
    data_comics_local = { "method" : {"type" : "comics"}, "mode" : "local", "run" : []}

    with open(comics_file,"r") as csvfile:
        result_dict = csv.DictReader(csvfile)
        for line in result_dict:
            if float(line["threshold"]) > max_thr:
                break
            if line["states"] == '-':
                els = { "threshold" : float(line["threshold"]), "value" : [-1], "wall_times" : [float(-1)]}
            else:
                els = { "threshold" : float(line["threshold"]), "value" : [int(line["states"])], "wall_times" : [float(line["t_wall"])]}
            if line["mode"] == "global":
                data_comics_global["run"].append(els)
            elif line["mode"] == "local":
                data_comics_local["run"].append(els)
            else:
                assert False

    return data_comics_global, data_comics_local


def alternative_init_plot(data_link,normalize=False,e_mode=False,ao_max_range=range(3),ao_min_range=range(3),invp_min_range=range(3),invf_max_range=range(3)):
    with open(data_link / "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)-max.json","r") as fp:
        data1 = json.load(fp)

    with open(data_link / "QSHeur(solver=gurobi,iterations=5,initializertype=InverseFrequencyInitializer,updatertype=InverseResultUpdater)-max.json","r") as fp:
        data2 = json.load(fp)

    with open(data_link / "QSHeur(solver=gurobi,iterations=5,initializertype=InverseReachabilityInitializer,updatertype=InverseResultUpdater)-max.json","r") as fp:
        data3 = json.load(fp)

    with open(data_link / "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)-min.json","r") as fp:
        data4 = json.load(fp)

    with open(data_link / "QSHeur(solver=gurobi,iterations=5,initializertype=InverseReachabilityInitializer,updatertype=InverseResultUpdater)-min.json","r") as fp:
        data5 = json.load(fp)

    with open(data_link / "QSHeur(solver=gurobi,iterations=5,initializertype=InverseFrequencyInitializer,updatertype=InverseResultUpdater)-min.json","r") as fp:
        data6 = json.load(fp)

    fig1, ax1 = plt.subplots()
    fig2, ax2 = plt.subplots()
    fig3, ax3 = plt.subplots()
    fig4, ax4 = plt.subplots()

    title = data_link.stem
    # title_max_states = data_link.stem.replace("_","-") + " max"
    # title_min_states = data_link.stem.replace("_","-") + " min"
    bm.render(data1, mode="states-thr",title=title, ax=ax1,normalize=normalize,sol_range=ao_max_range,custom_label="max-AO",e_mode=e_mode,plot_no=1)
    bm.render(data2, mode="states-thr",title=title, ax=ax1,normalize=normalize,sol_range=invf_max_range,custom_label="max-InvE",e_mode=e_mode,plot_no=2)
    #bm.render(data3, mode="states-thr",title=title, ax=ax1,normalize=normalize,sol_range=range(3),custom_label="max-InvP",e_mode=e_mode,plot_no=3)


    bm.render(data1, mode="wall_time-thr", ax=ax2, normalize=normalize,sol_range=ao_max_range,custom_label="max-AO",e_mode=e_mode,plot_no=1)
    bm.render(data2, mode="wall_time-thr", ax=ax2, normalize=normalize,sol_range=invf_max_range,custom_label="max-InvE",e_mode=e_mode,plot_no=2)

    bm.render(data4, mode="states-thr",title=title, ax=ax3, normalize=normalize,sol_range=ao_min_range,custom_label="min-AO",e_mode=e_mode,plot_no=1)
    bm.render(data5, mode="states-thr",title=title, ax=ax3, normalize=normalize,sol_range=invp_min_range,custom_label="min-InvP",e_mode=e_mode,plot_no=2)
    #bm.render(data6, mode="states-thr",title=title, ax=ax3,normalize=normalize,sol_range=range(3),custom_label="max-InvE",e_mode=e_mode,plot_no=3)

    bm.render(data4, mode="wall_time-thr", ax=ax4, normalize=normalize,sol_range=ao_min_range,custom_label="min-AO",e_mode=e_mode,plot_no=1)
    bm.render(data5, mode="wall_time-thr", ax=ax4, normalize=normalize,sol_range=invp_min_range,custom_label="min-InvP",e_mode=e_mode,plot_no=2)

    fig1.savefig("plots/" + data_link.stem + "_max_states.pdf")
    fig2.savefig("plots/" + data_link.stem + "_max_times.pdf")
    fig3.savefig("plots/" + data_link.stem + "_min_states.pdf")
    fig4.savefig("plots/" + data_link.stem + "_min_time.pdf")
    plt.close(fig1)
    plt.close(fig2)
    plt.close(fig3)
    plt.close(fig4)

### stats-tables

def create_stats_line(model,name,info_dict):
    if model == "DTMC":
        stats_line = "{0} & $\\num{{{1}}}$ & $\\num{{{2}}}$ \\\\ \n ".format(name,info_dict["nr_states"],info_dict["prmin"])
    elif model == "MDP":
        stats_line = "{0} & $\\num{{{1}}}$ & $\\num{{{2}}}$ & $\\num{{{3}}}$ & $\\num{{{4}}}$ \\\\ \n ".format(name,info_dict["nr_states"],info_dict["nr_sap"],info_dict["prmin"],info_dict["prmax"])
    else:
        assert 0
    return stats_line

# ## DTMC stats

crowds_instances = [("(2,6)","crowds/crowds_2_6/",False),("(2,8)","crowds/crowds_2_8/",False),("(5,8)","crowds/crowds_5_8/",False)]

brp_instances = [("(16,2)","brp/brp_16_2/",True),("(32,2)","brp/brp_32_2/",True),("(1024,2)","brp/brp_1024_2/",True)]

leader_instances = [("(6,6)","leader/leader_6_6/",False)] # ("(3,2)","leader/leader_3_2/",False)

all_instances = [("crowds", crowds_instances), ("brp",brp_instances), ("leader",leader_instances)]

no_instances = sum([len(i) for (n,i) in all_instances])
tab_dtmc_stats = "{\n"
tab_dtmc_stats += "\\begin{{tabular}}{{{0}}}\n ".format("c@{\hskip 0.5cm}ccc@{\hskip 0.5cm}ccc@{\hskip 0.5cm}c")

tab_dtmc_stats += "\\toprule "

##first-line header
for (name, instances) in all_instances:
        tab_dtmc_stats += "& \\multicolumn{{{0}}}{{c}}{{{1}}} ".format(len(instances),name)

tab_dtmc_stats += "\\\\ \n "

tab_dtmc_stats += "\\cmidrule(lr){2-4}\\cmidrule(lr){5-7}\\cmidrule(lr){8-8} \n"

##second-line header
for (name, instances) in all_instances:
    for (in_n, fol, sc_n) in instances:
            tab_dtmc_stats += " & \\multicolumn{{1}}{{c}}{{{0}}} ".format(in_n)


tab_dtmc_stats += "\\\\ \n "
tab_dtmc_stats += "\\midrule \n "

##states
tab_dtmc_stats += "\multicolumn{1}{c|}{states} "
first = True
for (name, instances) in all_instances:
    for (in_n, fol, sc_n) in instances:
        with open(data_folder / fol / "info.json") as fp:
            info_dict = json.load(fp)
        if first:
            tab_dtmc_stats += " & \hspace{{0.1cm}} $\\num{{{0}}}$ ".format(info_dict["nr_states"])
            first = False
        else:
            tab_dtmc_stats += " & $\\num{{{0}}}$ ".format(info_dict["nr_states"])

tab_dtmc_stats += "\\\\ \n "

##pr
tab_dtmc_stats += "\multicolumn{1}{c|}{$\\Pr(\\lozenge \\goal)$}"
first  =True
for (name, instances) in all_instances:
    for (in_n, fol, sc_n) in instances:
        with open(data_folder / fol / "info.json") as fp:
            info_dict = json.load(fp)
        if first:
            addspace = "\hspace{0.1cm}"
            first = False
        else:
            addspace = ""

        if sc_n == True:
            tab_dtmc_stats += " & {1} $\\num{{{0:.3e}}}$ ".format(info_dict["prmin"],addspace)
        else:
            tab_dtmc_stats += " & {1} $\\num{{{0:.3f}}}$ ".format(info_dict["prmin"],addspace)

tab_dtmc_stats += "\\\\ \n "
tab_dtmc_stats += "\\bottomrule "

tab_dtmc_stats += "\\end{tabular}\n}"

print(tab_dtmc_stats)

with open("tables/table_dtmc_stats.tex","w+") as fp:
    fp.write(tab_dtmc_stats)

## MDP stats

consensus_instances = [("(2,4)","consensus/consensus_2_4/",False),("(4,2)","consensus/consensus_4_2/",False),("(4,4)","consensus/consensus_4_2/",False)]

firewire_instances = [("3","firewire/firewire_3/",False),("30","firewire/firewire_30/",False)]

csma_instances = [("(2,6)","csma/csma_2_6/",False),("(3,2)","csma/csma_3_2/",False)]

all_instances = [("cons", consensus_instances), ("firewire",firewire_instances), ("csma",csma_instances)]

no_instances = sum([len(i) for (n,i) in all_instances])
tab_mdp_stats = "{\n"
tab_mdp_stats += "\\begin{{tabular}}{{{0}}}\n ".format("c@{\hskip 0.5cm}ccc@{\hskip 0.5cm}cc@{\hskip 0.5cm}cc")

tab_mdp_stats += "\\toprule "

##first-line header
for (name, instances) in all_instances:
    tab_mdp_stats += "& \\multicolumn{{{0}}}{{c}}{{{1}}} ".format(len(instances),name)

tab_mdp_stats += "\\\\ \n "

tab_mdp_stats += "\\cmidrule(lr){2-4}\\cmidrule(lr){5-6}\\cmidrule(lr){7-8} \n"

##second-line header
for (name, instances) in all_instances:
    for (in_n, fol, sc_n) in instances:
        tab_mdp_stats += " & \\multicolumn{{1}}{{c}}{{{0}}} ".format(in_n)

tab_mdp_stats += "\\\\ \n "
tab_mdp_stats += "\\midrule \n "

##states
tab_mdp_stats += "\multicolumn{1}{c|}{states} "
first = True
for (name, instances) in all_instances:
    for (in_n, fol, sc_n) in instances:
        with open(data_folder / fol / "info.json") as fp:
            info_dict = json.load(fp)
        if first:
            tab_mdp_stats += " & \hspace{{0.1cm}} $\\num{{{0}}}$ ".format(info_dict["nr_states"])
            first = False
        else:
            tab_mdp_stats += " & $\\num{{{0}}}$ ".format(info_dict["nr_states"])

tab_mdp_stats += "\\\\ \n "

##states
tab_mdp_stats += "\multicolumn{1}{c|}{state-action pairs}"
first = True
for (name, instances) in all_instances:
    for (in_n, fol, sc_n) in instances:
        with open(data_folder / fol / "info.json") as fp:
            info_dict = json.load(fp)
        if first:
            tab_mdp_stats += " & \hspace{{0.1cm}} $\\num{{{0}}}$ ".format(info_dict["nr_sap"])
            first = False
        else:
            tab_mdp_stats += " & $\\num{{{0}}}$ ".format(info_dict["nr_sap"])

tab_mdp_stats += "\\\\ \n "

# ##prmin
# tab_mdp_stats += "$\\Pr^{\\min}(\\lozenge \\goal)$"
# for (name, instances) in all_instances:
#     for (in_n, fol, sc_n) in instances:
#         with open(data_folder / fol / "info.json") as fp:
#             info_dict = json.load(fp)
#         if sc_n == True:
#             tab_mdp_stats += " & $\\num{{{0:.3e}}}$ ".format(info_dict["prmin"])
#         else:
#             tab_mdp_stats += " & $\\num{{{0:.3f}}}$ ".format(info_dict["prmin"])

# tab_mdp_stats += "\\\\ \n "

# ##prmax
# tab_mdp_stats += "$\\Pr^{\\max}(\\lozenge \\goal)$"
# for (name, instances) in all_instances:
#     for (in_n, fol, sc_n) in instances:
#         with open(data_folder / fol / "info.json") as fp:
#             info_dict = json.load(fp)
#         if sc_n == True:
#             tab_mdp_stats += " & $\\num{{{0:.3e}}}$ ".format(info_dict["prmax"])
#         else:
#             tab_mdp_stats += " & $\\num{{{0:.3f}}}$ ".format(info_dict["prmax"])

tab_mdp_stats += "\\bottomrule "

tab_mdp_stats += "\\end{tabular}\n}"

print(tab_mdp_stats)

with open("tables/table_mdp_stats.tex","w+") as fp:
    fp.write(tab_mdp_stats)



### Compute minimal witnesses (MILP)

### Table1

configs = [("DTMC","crowds_2_8","crowds/crowds_2_8/",[0.05,0.11,0.33,0.4]),
           ("DTMC","brp_32_2","brp/brp_32_2",[2e-6,8e-6,1e-5,1.2e-5]),
           ("MDP","consensus_2_4","consensus/consensus_2_4",[0.1,0.3,0.6,0.9]),
           ("MDP","firewire_3","firewire/firewire_3",[0.1,0.3,0.6,0.9])]

tab_1_str =  "{\n"
tab_1_str += "\\begin{tabular}{l l l l l l}\n"

first_mdp = True

for (model_type,name,directory,thr_list) in configs:
    data_min_file = data_folder / directory / "MILPExact(solver=gurobi)-min.json"
    data_max_file = data_folder / directory / "MILPExact(solver=gurobi)-max.json"

    with open(data_min_file,"r") as fp:
        data_min = json.load(fp)

    with open(data_max_file,"r") as fp:
        data_max = json.load(fp)

    datasets = [("$\min$-MILP",data_min),("& $\max$-MILP",data_max)]

    if model_type == "MDP":
        data_ltlsubsys_file = data_folder / directory / "ltlsubsys-run.json"
        with open(data_ltlsubsys_file,"r") as fp:
            data_ltlsubsys = json.load(fp)

        datasets.append(("& ltlsubsys ($\max$)",data_ltlsubsys))
        no_rows = 3
    else:
        no_rows = 2

    tab_1_str += " & \multicolumn{0}{c}{$\lambda =$}"
    for thr in thr_list:
        tab_1_str += "& \\multicolumn{{0}}{{c}}{{$\\num{{{0}}}$}} ".format(thr)
    tab_1_str += "\\\\ \hline \\noalign{\\vskip 0.1cm} \n "

    tab_1_str += "\\multirow{{{0}}}*{{{1}}} & ".format(no_rows,re.sub("_","\_",name))

    i = 1
    for data_set_name, data in datasets:

        tab_1_str += data_set_name
        for thr in thr_list:
            found = False
            for d in data["run"]:
                if math.isclose(d["threshold"],thr):
                    if isinstance(d["value"],list):
                        val = d["value"][0]
                        time = d["wall_times"][0]
                    else:
                        val = d["value"]
                        time = d["wall_times"]

                    if val == -1:
                        tab_1_str += "& TO (-)"
                    elif time < 1:
                        tab_1_str += " & $<1$ ({1})".format(time,int(val))
                    else:
                        tab_1_str += " & {0:.1f} ({1})".format(time,int(val))
                    found = True
                    break
            if not found:
                tab_1_str += "& TO (-)"

        if i == len(datasets):
            tab_1_str += "\\\\[0.2cm] \n "
        else:
            i += 1
            tab_1_str += "\\\\ \n "
tab_1_str += "\\end{tabular}\n}"

print(tab_1_str)

with open("tables/table1.tex","w+") as fp:
    fp.write(tab_1_str)


print("1")
#QSHeur eval
#Figure 1 - iterations / spikes / min-max in DTMCs

datasets = [("crowds_2_8","crowds/crowds_2_8",False,False), ("brp_1024_2","brp/brp_1024_2",True,True)]

qsheur_name = "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)"

for (name,folder,normalize,emode) in datasets:

    with open(data_folder / folder / Path(qsheur_name + "-min.json"),"r") as fp:
        data_min = json.load(fp)
    with open(data_folder / folder / Path(qsheur_name + "-max.json"),"r") as fp:
        data_max = json.load(fp)

    fig, ax = plt.subplots()
    bm.render(data_min, mode="states-thr",title=name, ax=ax,normalize=normalize,sol_range=range(5),custom_label= "min-QSHeur",e_mode=emode)
    fig.savefig("plots/" + "fig_qsheur-" + name + "-min.pdf")

    fig, ax = plt.subplots()
    bm.render(data_max, mode="states-thr",title=name, ax=ax,normalize=normalize,sol_range=range(5),custom_label="max-QSHeur",e_mode=emode)
    fig.savefig("plots/" + "fig_qsheur-" + name + "-max.pdf")
    plt.close(fig)

print("2")
## Figure 2 - comparison with exact methods

datasets = [("crowds_2_6","crowds/crowds_2_6",False),
            ("crowds_2_8","crowds/crowds_2_8",False),
            #("brp_8_2","brp/brp_8_2",True),
            ("brp_32_2","brp/brp_32_2",True),
            ("brp_16_2","brp/brp_16_2",True)]

qsheur_name = "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)"

exact_name = "MILPExact(solver=gurobi)-min"

for (name,folder,emode) in datasets:

    with open(data_folder / folder / Path(qsheur_name + "-min.json"),"r") as fp:
        data_heur_min = json.load(fp)
    with open(data_folder / folder / Path(qsheur_name + "-max.json"),"r") as fp:
        data_heur_max = json.load(fp)

    with open(data_folder / folder / Path(exact_name + ".json"),"r") as fp:
        data_exact = json.load(fp)

    fig, ax = plt.subplots()
    bm.render(data_heur_max, mode="states-thr",title=name,ax=ax,normalize=False,sol_range=[2],custom_label= "max-QSHeur",e_mode=emode,plot_no=1)

    bm.render(data_exact, mode="states-thr", ax=ax,normalize=False,e_mode=emode,plot_no=3)

    bm.render(data_heur_min, mode="states-thr",title=name,ax=ax,normalize=False,sol_range=[2],custom_label= "min-QSHeur",e_mode=emode,plot_no=2)

    fig.savefig("plots/" + "fig_qsheur-vs-exact-" + name + ".pdf")
    plt.close(fig)


print("3")
## Figure 3 - comparison with comics

datasets = [("crowds_5_8","crowds/crowds_5_8",False), ("brp_1024_2","brp/brp_1024_2",True)]

qsheur_name = "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)"

for (name,folder,emode) in datasets:

    with open(data_folder / folder / Path(qsheur_name + "-min.json"),"r") as fp:
        data_heur_min = json.load(fp)
    with open(data_folder / folder / Path(qsheur_name + "-max.json"),"r") as fp:
        data_heur_max = json.load(fp)

    with open(data_folder / folder / Path("info.json"),"r") as fp:
        info_dict = json.load(fp)

    data_gl, data_loc = comics_to_proper_json(data_folder / folder / "comics-run.csv", info_dict["prmin"])
    # print(info_dict["prmin"])
    # print(data_gl)
    # print(data_loc)
    
    #compute a value to use for timeouts
    max_subsys_size = 0
    for d in [data_heur_min,data_heur_max,data_gl]:  # add data_loc
        max_subsys_size = max([max_subsys_size] + [max([max(el["value"]) for el in d["run"]])])

    timeout_val = (max_subsys_size + max_subsys_size * 0.02)

    fig, ax = plt.subplots()

    bm.render(data_heur_max, mode="states-thr",title=name,ax=ax,normalize=True,sol_range=[2],custom_label= "max-QSHeur",e_mode=emode,markers=["o"],timeout_val=timeout_val)

    bm.render(data_heur_min, mode="states-thr",title=name,ax=ax,normalize=True,sol_range=[2],custom_label= "min-QSHeur",e_mode=emode,markers=["^"],timeout_val=timeout_val)

    bm.render(data_gl, mode="states-thr",title=name,ax=ax,normalize=True,custom_label= "comics-global",e_mode=emode,markers=["s"],timeout_val=timeout_val)

    if len(data_loc["run"]) != 0:
        bm.render(data_loc, mode="states-thr",title=name,ax=ax,normalize=True,custom_label= "comics-local",e_mode=emode,markers=["x"],timeout_val=timeout_val)

    ax2 = ax.twinx()
    ax2.set_ylim(ax.get_ylim())
    ax2.set_yticks([timeout_val/1000])
    ax2.set_yticklabels(["TO"])
    ax2.axhline(y = timeout_val/1000,linestyle="dotted")

    if name == "crowds_5_8":
        ax.legend(bbox_to_anchor=(0,0.95),loc='upper left')


    fig.savefig("plots/" + "fig_qsheur-vs-comics-states-" + name + ".pdf")
    plt.close(fig)

    ## plots for computation times

    #compute a value to use for timeouts
    max_time = 0
    for d in [data_heur_min,data_heur_max,data_gl,data_loc]:
        max_time = max([max_time] + [max([max(el["wall_times"]) for el in d["run"]])])

    timeout_val = (max_time + max_time * 0.02)

    fig, ax = plt.subplots()

    ax.set_yscale('log')

    bm.render(data_heur_max, mode="wall_time-thr",title=name,ax=ax,normalize=False,sol_range=[2],custom_label= "max-QSHeur",e_mode=emode,markers=["o"],timeout_val=timeout_val)

    bm.render(data_heur_min, mode="wall_time-thr",title=name,ax=ax,normalize=False,sol_range=[2],custom_label= "min-QSHeur",e_mode=emode,markers=["^"],timeout_val=timeout_val)

    bm.render(data_gl, mode="wall_time-thr",title=name,ax=ax,normalize=False,custom_label= "comics-global",e_mode=emode,markers=["s"],timeout_val=timeout_val)

    if len(data_loc["run"]) != 0:
        bm.render(data_loc, mode="wall_time-thr",title=name,ax=ax,normalize=False,custom_label= "comics-local",e_mode=emode,markers=["x"],timeout_val=timeout_val)

    ax2 = ax.twinx()
    ax2.set_yscale('log')
    ax2.set_ylim(ax.get_ylim())
    ax2.set_yticks([timeout_val])
    ax2.set_yticklabels(["TO"])
    ax2.axhline(y = timeout_val,linestyle="dotted")

    if name == "crowds_5_8":
        ax.legend(bbox_to_anchor=(0,0.95),loc='upper left')

    fig.savefig("plots/" + "fig_qsheur-vs-comics-times-" + name + ".pdf")
    plt.close(fig)

print("4")
## Fig 4 - alternative initialization objectives

data_sets = [(Path("crowds/crowds_5_8"),True,False,range(2),range(1),range(1),range(1)),
             (Path("brp/brp_1024_2"),True,True,range(3),range(3),range(3),range(3)),
             (Path("csma/csma_2_6"),True,False,range(3),range(1),range(1),range(1)),
#             (Path("consensus/consensus_2_16"),False,False,range(3),range(3),range(3),range(3)),
             (Path("consensus/consensus_4_2"),True,False,range(2),range(3),range(3),range(1))]

for (name,normalize,emode,ao_max_range,ao_min_range,invp_min_range,invf_max_range) in data_sets:
    alternative_init_plot(data_folder / name,normalize=normalize,e_mode=emode,ao_max_range=ao_max_range,ao_min_range=ao_min_range,invp_min_range=invp_min_range,invf_max_range=invf_max_range)

print("5")
## Fig 5 - label-based minimization

data_sets = [("brp_32_8","brp_labeled/brp.32.8",False),
             ("brp_1024_8","brp_labeled/brp.1024.8",True)]

qsheur_name = "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)"

exact_name = "MILPExact(solver=gurobi)"

for (name,folder,box_special) in data_sets:

    with open(data_folder / folder / Path(qsheur_name + "-min.json"),"r") as fp:
        data_heur_min = json.load(fp)
    with open(data_folder / folder / Path(qsheur_name + "-max.json"),"r") as fp:
        data_heur_max = json.load(fp)
    with open(data_folder / folder / Path(exact_name + "-min.json"),"r") as fp:
        data_ex_min = json.load(fp)
    with open(data_folder / folder / Path(exact_name + "-max.json"),"r") as fp:
        data_ex_max = json.load(fp)

    timeout_val = 10

    all_thr = [ el["threshold"] for el in data_ex_min["run"] ]
    data_heur_max = add_timeouts(all_thr,data_heur_max)
    data_ex_max = add_timeouts(all_thr,data_ex_max)

    data_heur_max["run"] = [ el for el in data_heur_max["run"] if not math.isclose(el["threshold"],1) ]
    data_ex_max["run"] = [ el for el in data_ex_max["run"] if not math.isclose(el["threshold"],1) ]

    fig, ax = plt.subplots()

    bm.render(data_ex_min, mode="states-thr",title=name,ax=ax,normalize=False,custom_label= "MILPExact",e_mode=False,markers=["s"],timeout_val=timeout_val)

    bm.render(data_heur_max, mode="states-thr",title=name,ax=ax,normalize=False,sol_range=range(2),custom_label= "max-QSHeur",e_mode=False,markers=["o"],timeout_val=timeout_val)

    bm.render(data_heur_min, mode="states-thr",title=name,ax=ax,normalize=False,sol_range=range(1),custom_label= "min-QSHeur",e_mode=False,markers=["^"],timeout_val=timeout_val)

    ax2 = ax.twinx()
    ax2.set_ylim(ax.get_ylim())
    ax2.set_yticks([timeout_val])
    ax2.set_yticklabels(["TO"])
    ax2.axhline(y = timeout_val,linestyle="dotted")

    fig.savefig("plots/" + "fig_label-based-states-" + name + ".pdf")

    ## plots for computation times

    #compute a value to use for timeouts
    max_time = 0
    for d in [data_heur_min,data_heur_max,data_ex_min,data_ex_max]:
        max_time = max([max_time] + [max([max(el["wall_times"]) for el in d["run"]])])

    timeout_val = (max_time + max_time * 0.2)

    fig, ax = plt.subplots()

    ax.set_yscale('log')

    bm.render(data_heur_max, mode="wall_time-thr",title=name,ax=ax,normalize=False,sol_range=range(2),custom_label= "max-QSHeur",markers=["o"],timeout_val=timeout_val)

    bm.render(data_heur_min, mode="wall_time-thr",title=name,ax=ax,normalize=False,sol_range=range(2),custom_label= "min-QSHeur",markers=["^"],timeout_val=timeout_val)

    bm.render(data_ex_max, mode="wall_time-thr",title=name,ax=ax,normalize=False,custom_label= "MILPExact-max",markers=["s"],timeout_val=timeout_val)

    bm.render(data_ex_min, mode="wall_time-thr",title=name,ax=ax,normalize=False,custom_label= "MILPExact-min",markers=["x"],timeout_val=timeout_val)

    if box_special:
        ax.set_ylim((10,ax.get_ylim()[1]))
        ax.legend(bbox_to_anchor=(1,0),loc='lower right')

    ax2 = ax.twinx()
    ax2.set_yscale('log')
    ax2.set_ylim(ax.get_ylim())
    ax2.set_yticks([timeout_val])
    ax2.set_yticklabels(["TO"])
    ax2.axhline(y = timeout_val,linestyle="dotted")

    fig.savefig("plots/" + "fig_label-based-times-" + name + ".pdf",bbox_inches='tight')
    plt.close(fig)


# table for label-based

tab_2_str =  "{\n"
tab_2_str += "\\begin{tabular}{l l c c c c c}\n"

iteration = 3

#lambdas
thr_list = [0.97,0.978,0.982, 0.986,0.994]

tab_2_str += " & $\lambda =$"
for thr in thr_list:
    tab_2_str += "& $\\num{{{0}}} $ ".format(thr)

for (name,directory,_) in data_sets:
    data_heur_min_file = data_folder / directory / Path(qsheur_name + "-min.json")
    data_heur_max_file = data_folder / directory / Path(qsheur_name + "-max.json")
    data_ex_file = data_folder / directory / "MILPExact(solver=gurobi)-min.json"

    with open(data_heur_min_file,"r") as fp:
        data_heur_min = json.load(fp)
    with open(data_heur_max_file,"r") as fp:
        data_heur_max = json.load(fp)
    with open(data_ex_file,"r") as fp:
        data_ex = json.load(fp)

    print(data_heur_max)

    instance_sets = [("min-QSHeur$_{0}$".format(iteration), data_heur_min),
                     ("max-QSHeur$_{0}$".format(iteration), data_heur_max),
                     ("exact", data_ex) ]

    tab_2_str += "\\\\[0.02cm] \hline \\\\[0.02cm] \n "

    tab_2_str += "\\multirow{{{0}}}*{{{1}}} & ".format(3,re.sub("_","\_",name))

    first = True
    for data_set_name, data in instance_sets:

        if first == True:
            tab_2_str += data_set_name
            first = False
        else:
            tab_2_str += "\\\\[0.02cm] \n"
            tab_2_str +=  "& " + data_set_name

        for thr in thr_list:
            found = False
            for d in data["run"]:
                if math.isclose(d["threshold"],thr):
                    if data_set_name == "exact":
                        tab_2_str += " & {0}".format(int(d["value"][0]))
                        found = True
                        break
                    elif len(d["value"]) < iteration or d["value"][iteration] == -1:
                        break
                    else:
                        tab_2_str += " & {0}".format(int(d["value"][iteration]))
                        found = True
                        break
            if not found:
                tab_2_str += "& TO"

tab_2_str += "\\\\ \\end{tabular}\n}"

print(tab_2_str)

with open("tables/table2.tex","w+") as fp:
    fp.write(tab_2_str)


## global table / table-3

qsheur_ao = "QSHeur(solver=gurobi,iterations=5,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)"
qsheur_ao_threeit = "QSHeur(solver=gurobi,iterations=3,initializertype=AllOnesInitializer,updatertype=InverseResultUpdater)"
qsheur_invf = "QSHeur(solver=gurobi,iterations=5,initializertype=InverseFrequencyInitializer,updatertype=InverseResultUpdater)"
qsheur_invp = "QSHeur(solver=gurobi,iterations=5,initializertype=InverseReachabilityInitializer,updatertype=InverseResultUpdater)"

nr_thrs = 4
data_sets = [("brp_32_2","brp/brp_32_2",[2e-6,1e-5,1.8e-5,2e-5],False,"DTMC"),
             ("brp_1024_2","brp/brp_1024_2",[2e-6,1e-5,1.8e-5,2e-5],False,"DTMC"),
             ("crowds_2_8","crowds/crowds_2_8",[0.05,0.15,0.21,0.29],False,"DTMC"),
             ("crowds_5_8","crowds/crowds_5_8",[0.05,0.15,0.21,0.29],False,"DTMC"),
             ("leader_6_6","leader/leader_6_6",[0.1,0.3,0.7,0.9],False,"DTMC"),
             ("csma_2_6","csma/csma_2_6",[0.1,0.3,0.6,0.9],False,"MDP"),
             ("csma_3_2","csma/csma_3_2",[0.1,0.3,0.6,0.9],False,"MDP"),
             ("cons_2_4","consensus/consensus_2_4",[0.1,0.3,0.6,0.9],False,"MDP"),
             ("cons_4_4","consensus/consensus_4_4",[0.1,0.3,0.6,0.9],False,"MDP"),
             ("firewire_3","firewire/firewire_3",[0.1,0.3,0.6,0.9],False,"MDP"),
             ("firewire_30","firewire/firewire_30",[0.1,0.3,0.6,0.9],False,"MDP"),
             ]

lambda_sets = [("brp",data_sets[0][2],True),("crowds",data_sets[2][2],False),("leader",data_sets[4][2],False),
               ("csma",data_sets[5][2],False),("cons",data_sets[7][2],False),("firewire",data_sets[-1][2],False)]


for m_type in ["DTMC","MDP"]:
    tab_3_str =  "{\n"
    tab_3_str += "\\begin{tabular}{l l l l l l l}\n"
    
    iteration = 3
    
    tab_3_str += " & "
    
    for i in range(nr_thrs):
        tab_3_str += "& \\multicolumn{{1}}{{c}}{{$\lambda_{{{0}}}$ }} ".format(i+1)
    
    for (name,directory,thrlist,include_exact,modeltype) in data_sets:
        if modeltype != m_type:
            continue
        info_file = data_folder / directory / Path("info.json")
        data_heur_min_ao_file = data_folder / directory / Path(qsheur_ao + "-min.json")
        data_heur_max_ao_file = data_folder / directory / Path(qsheur_ao + "-max.json")
        data_heur_min_invp_file = data_folder / directory / Path(qsheur_invp + "-min.json")
        data_heur_max_invf_file = data_folder / directory / Path(qsheur_invf + "-max.json")
        data_ex_min_file = data_folder / directory / "MILPExact(solver=gurobi)-min.json"
        data_ex_max_file = data_folder / directory / "MILPExact(solver=gurobi)-min.json"
    
        with open(data_heur_min_ao_file,"r") as fp:
            data_heur_min_ao = json.load(fp)
        with open(data_heur_max_ao_file,"r") as fp:
            data_heur_max_ao = json.load(fp)
        with open(data_heur_min_invp_file,"r") as fp:
            data_heur_min_invp = json.load(fp)
        with open(data_heur_max_invf_file,"r") as fp:
            data_heur_max_invf = json.load(fp)
        with open(info_file,"r") as fp:
            info_dict = json.load(fp)
    
        instance_sets = [("min-qs-ao".format(iteration), "qs", data_heur_min_ao),
                         ("min-qs-invp".format(iteration), "qs", data_heur_min_invp),
                         ("max-qs-ao".format(iteration), "qs", data_heur_max_ao),
                         ("max-qs-invf".format(iteration), "qs", data_heur_max_invf)]
    
        if include_exact:
            with open(data_ex_min_file,"r") as fp:
                data_exmin = json.load(fp)
            with open(data_ex_max_file,"r") as fp:
                data_exmax = json.load(fp)
            instance_sets.append(("min-MILP", "exact", data_ex_min))
            instance_sets.append(("max-MILP", "exact", data_ex_max))
    
        if modeltype == "DTMC":
            data_comics_gl, data_comics_loc = comics_to_proper_json(data_folder / directory / "comics-run.csv",info_dict["prmin"])
            instance_sets.append(("comics-gl", "comics", data_comics_gl))
            instance_sets.append(("comics-lo", "comics", data_comics_loc))
    
        #tab_3_str += "\\\\[0.02cm] \hline \\\\[0.02cm] \n "
        tab_3_str += "\\\\ \\hline \n "
    
        with open(data_folder / directory / "info.json","r") as fp:
            info_json = json.load(fp)
            nr_states = info_json["nr_states"]
            if modeltype == "MDP":
                nr_sap = info_json["nr_sap"]
    
        if modeltype == "DTMC":
            tab_3_str += "\\multirow{{{0}}}*{{ \\shortstack{{ {1} \\\\ {{ \\scriptsize $|S| = \\num{{ {2} }}$ }} }} }} & ".format(len(instance_sets),re.sub("_","\_",name),nr_states)
        elif modeltype == "MDP":
            tab_3_str += "\\multirow{{{0}}}*{{ \\shortstack{{ {1} \\\\ {{ \\scriptsize $|S| = \\num{{ {2} }}$ }} \\\\ {{ \\scriptsize $|\enabled| = \\num{{ {3} }}$ }} }} }} & ".format(len(instance_sets),re.sub("_","\_",name),nr_states,nr_sap)
    
        first = True
        for data_set_name, data_set_type, data in instance_sets:
    
            if first == True:
                tab_3_str += data_set_name
                first = False
            else:
                tab_3_str += "\\\\ \n"
                tab_3_str +=  "& " + data_set_name
    
            for thr in thrlist:
                found = False
                for d in data["run"]:
                    if math.isclose(d["threshold"],thr):
                        found = True
                        if data_set_type == "exact":
                            if d["wall_times"][0] >= 1:
                                tab_3_str += " & $\\num{{{0:.1f}}}$ ($\\num{{{1}}}$)".format(d["wall_times"][0],int(d["value"][0]))
                            else: 
                                tab_3_str += " & $<1$ ($\\num{{{0}}}$)".format(int(d["value"][0]))
                            break
                        elif data_set_type == "comics":
                            if d["value"][0] == -1:
                                tab_3_str += "& -"
                                break
                            if d["wall_times"][0] >= 1:
                                tab_3_str += " & $\\num{{{0:.1f}}}$ ($\\num{{{1}}}$)".format(d["wall_times"][0],int(d["value"][0]))
                            else: 
                                tab_3_str += " & $<1$ ($\\num{{{0}}}$)".format(int(d["value"][0]))
                        elif len(d["value"]) < iteration or d["value"][iteration-1] == -1:
                            tab_3_str += "& -"
                            break
                        else:
                            if d["wall_times"][iteration-1] >= 1200:
                                tab_3_str += "& -"
                            elif d["wall_times"][iteration-1] >= 1:
                                tab_3_str += " & $\\num{{{0:.1f}}}$ ($\\num{{{1}}}$)".format(d["wall_times"][iteration-1],int(d["value"][iteration-1]))
                            else: 
                                tab_3_str += " & $<1$ ($\\num{{{0}}}$)".format(int(d["value"][iteration-1]))
                            break
                if not found:
                    tab_3_str += "& -"
    
    tab_3_str += "\\\\ \\end{tabular}\n}"
    
    print(tab_3_str)
    
    with open("tables/table-full-heuristics-{0}.tex".format(m_type),"w+") as fp:
        fp.write(tab_3_str)
    tab_3_str = ""
    

## lambdas-table


tab_lambdas = "{\n"
tab_lambdas += "\\begin{tabular}{lllll}\n"

tab_lambdas += "& $\lambda_1$ & $\lambda_2$ & $\lambda_3$ & $\lambda_4$ \\\\ \n"
tab_lambdas += "\\midrule \n"

for (name,ls,sc_n) in lambda_sets:
    if sc_n:
        tab_lambdas += "{0} & $\\num{{{1:.1e}}}$ & $\\num{{{2:.1e}}}$ & $\\num{{{3:.1e}}}$ & $\\num{{{4:.1e}}}$ \\\\ \n".format(name,ls[0],ls[1],ls[2],ls[3])
    else:
        tab_lambdas += "{0} & $\\num{{{1}}}$ & $\\num{{{2}}}$ & $\\num{{{3}}}$ & $\\num{{{4}}}$ \\\\ \n".format(name,ls[0],ls[1],ls[2],ls[3])
tab_lambdas += "\\\\ \\end{tabular}\n}"

with open("tables/table_lambdas.tex","w+") as fp:
    fp.write(tab_lambdas)

## Figure for tree-like

N_vals = [10,20,30,40,50,60,70,80,90]

threshold_nos = [0,4,6]

plt.rc('axes', labelsize=13)

timeout_val = 20*60

for MAX in [1]:
    for threshold_no in threshold_nos:
        gurobi_times_max = []
        gurobi_times_min = []
        tree_algo_times = []
        fig,ax = plt.subplots()
        for N in N_vals:
            print(str(N) + str(" ") + str(threshold_no))
            with open(data_folder / Path("tree-algo/brp.%d.%d/MILPExact(solver=gurobi)-max.json" % (N,MAX))) as fp:
                data_gurobi_max = json.load(fp)
                if len(data_gurobi_max["run"]) <= threshold_no:
                    gurobi_times_max.append(timeout_val)
                else:
                    gurobi_times_max.append(data_gurobi_max["run"][threshold_no]["wall_times"][0])
            with open(data_folder / Path("tree-algo/brp.%d.%d/MILPExact(solver=gurobi)-min.json" % (N,MAX))) as fp:
                data_gurobi_min = json.load(fp)
                if len(data_gurobi_min["run"]) <= threshold_no:
                    gurobi_times_min.append(timeout_val)
                else:
                    gurobi_times_min.append(data_gurobi_min["run"][threshold_no]["wall_times"][0])
            with open(data_folder / Path("tree-algo/brp.%d.%d/TreeAlgo(known_upper_bound=None,tol=1e-12)-max.json" % (N,MAX))) as fp:
                data_tree = json.load(fp)
                if len(data_tree["run"]) <= threshold_no:
                    tree_algo_times.append(timeout_val)
                else:
                    thr = data_tree["run"][threshold_no]["threshold"]
                    tree_algo_times.append(data_tree["run"][threshold_no]["wall_times"][0])

        plt.yscale("log")
        ax.set_xlabel("brp_N_1, threshold: {:.2}".format(thr))
        ax.set_ylabel("time in seconds".format(timeout_val))
        ax.plot(N_vals,gurobi_times_max, linestyle="dashed", marker="o",label="max-MILP")
        ax.plot(N_vals,gurobi_times_min, linestyle="dashed", marker="x",label="min-MILP")
        ax.plot(N_vals,tree_algo_times, linestyle="dashdot", marker="s",label="tree-algo")
        ax.grid(True)
        ax.legend()

        ax2 = ax.twinx()
        ax2.set_yscale('log')
        ax2.set_ylim(ax.get_ylim())
        ax2.set_yticks([timeout_val])
        ax2.set_yticklabels(["TO"])
        ax2.axhline(y = timeout_val,linestyle="dotted")

        if threshold_no == 0:
            ax.legend(bbox_to_anchor=(0,0.95),loc='upper left')

        fig.savefig("plots/tree-algo-plot_brp_X_{:.2}.pdf".format(thr))

        plt.close(fig)



assert 0
